//--------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//--------------------------------------------------------------------------
//
//	VectorFunctors.cc --- implementation of vector related operator classes
//
//	DESCRIPTION:
//		Defines of functions bound to the operators recognized
//		by the expression scanner. For a list of operators/functions,
//		see Selector.h
//
//	CHANGE HISTORY:
//		UL, 2002-05-23: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers (UL), Thu May 23 14:06:29 MEST 2002
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/selcore/VectorFunctors.cc,v 1.3 2009/06/26 15:51:59 aibarra Exp $
//
#include <cmath>

#include "basicdefs.h"
#include "VectorFunctors.h"
#include <iomanip>
#include <numeric>

using namespace SelectError;

//
//	constructors/destructors for operator classes
//
Vector::Vector(OpNode x, OpNode y, OpNode z)
{
	if (x->isConst() && y->isConst() && z->isConst()) {
		//	all components are const - d/f eval method is ok
		double *v = data();
		v[0] = x->arith(x);
		v[1] = y->arith(y);
		v[2] = z->arith(z);
	} else {
		OpNode v[] = { x, y, z };
		addChildren(NUM_OF(v), v);
		eval = &VectorNode::evalVariable;
	}
} // Vector::Vector()


CartesianVector &Vector::evalVariable() { return *this; }


UnitVector::UnitVector(OpNode x, OpNode y, OpNode z)
{
	if (x->isConst() && y->isConst() && z->isConst()) {
		//	all components are const
		double *v = data();
		v[0] = x->arith(x);
		v[1] = y->arith(y);
		v[2] = z->arith(z);
		normalize();
	} else {
		OpNode v[] = { x, y, z };
		addChildren(NUM_OF(v), v);
		eval = &VectorNode::evalVariable;
	}
} // UnitVector::UnitVector()


CartesianVector &UnitVector::evalVariable()
{
	standardEvalVariable();
	return normalize();
} // UnitVector::evalVariable();


SkyVector::SkyVector(OpNode ra, OpNode dec)
{
	if (ra->isConst() && dec->isConst()) {
		EquatorialDirection ed(RightAscension(ra->arith(ra)),
							   Declination(dec->arith(dec)));
		CartesianVector *cv = this;
		*cv = ed.toUnitCartesian();
	} else {
		OpNode v[] = { ra, dec };
		addChildren(NUM_OF(v), v);
		eval = &VectorNode::evalVariable;
	}
} // SkyVector::SkyVector()


CartesianVector &SkyVector::evalVariable()
{
	OpNode c0 = child(0),
		   c1 = child(1);
	EquatorialDirection ed(RightAscension(c0->arith(c0)),
						   Declination(c1->arith(c1)));
	CartesianVector *cv = this;
	return *cv = ed.toUnitCartesian();
} // SkyVector::evalVariable()


VectorAdd::VectorAdd(OpNode v1, OpNode v2)
{
	OpNode v[] = { v1, v2 };
	addChildren(NUM_OF(v), v);
	eval = &VectorNode::evalVariable;
} // VectorAdd::VectorAdd()


CartesianVector &VectorAdd::evalVariable()
{
	VectorNode *v1 = (VectorNode *)child(0),
			   *v2 = (VectorNode *)child(1);

	CartesianVector *v = this;
	return *v = (v1->*v1->eval)() + (v2->*v2->eval)();
} // VectorAdd::evalVariable()


VectorSub::VectorSub(OpNode v1, OpNode v2)
{
	OpNode v[] = { v1, v2 };
	addChildren(NUM_OF(v), v);
	eval = &VectorNode::evalVariable;
} // VectorSub::VectorSub()


CartesianVector &VectorSub::evalVariable()
{
	VectorNode *v1 = (VectorNode *)child(0),
			   *v2 = (VectorNode *)child(1);

	CartesianVector *v = this;
	return *v = (v1->*v1->eval)() - (v2->*v2->eval)();
} // VectorSub::evalVariable()


VectorScalarMult::VectorScalarMult(OpNode v1, OpNode v2)
{
	OpNode v[] = { v1, v2 };
	addChildren(NUM_OF(v), v);
	eval = &VectorNode::evalVariable;
} // VectorScalarMult::VectorScalarMult()


CartesianVector &VectorScalarMult::evalVariable()
{
	VectorNode *v1 = (VectorNode *)child(0);
	double	   c = child(1)->arith(child(1));

	CartesianVector *v = this;
	return *v = c * (v1->*v1->eval)();
} // VectorScalarMult::evalVariable()


VectorScalarDiv::VectorScalarDiv(OpNode v1, OpNode v2)
{
	OpNode v[] = { v1, v2 };
	addChildren(NUM_OF(v), v);
	eval = &VectorNode::evalVariable;
} // VectorScalarDiv::VectorScalarDiv()


CartesianVector &VectorScalarDiv::evalVariable()
{
	VectorNode *v1 = (VectorNode *)child(0);
	double	   c = child(1)->arith(child(1));

	CartesianVector *v = this;
	return *v = (v1->*v1->eval)()/c;
} // VectorScalarDiv::evalVariable()


CrossProd::CrossProd(OpNode v1, OpNode v2)
{
	OpNode v[] = { v1, v2 };
	addChildren(NUM_OF(v), v);
	eval = &VectorNode::evalVariable;
} // CrossProd::CrossProd()


CartesianVector &CrossProd::evalVariable()
{
	VectorNode *v1 = (VectorNode *)child(0),
			   *v2 = (VectorNode *)child(1);

	CartesianVector *v = this;
	return *v = crossProd((v1->*v1->eval)(), (v2->*v2->eval)());
} // CrossProd::evalVariable()


VectorMinus::VectorMinus(OpNode vec)
{
	OpNode v[] = { vec };
	addChildren(NUM_OF(v), v);
	eval = &VectorNode::evalVariable;
} // VectorMinus::VectorMinus()


CartesianVector &VectorMinus::evalVariable()
{
	VectorNode *v = (VectorNode *)child(0);

	CartesianVector *vec = this;
	return *vec = (v->*v->eval)();
} // VectorMinus::evalVariable()


#define OPCLASS2_A(op)												\
op::op() { registerArithFunc( op##Arith ); }						\
op::op(const OpNode n1, const OpNode n2) {							\
	OpNode nv[] = { n1, n2 }; addChildren(2, nv);					\
registerArithFunc( op##Arith ); }									\
op::~op() {}

static double VectorComponentArith(OpNode me)
{
	unsigned i = UNT(me->child(1)->arith(me->child(1)));
	if (i>2)
		selerrstr<<"Cannot access vector element "<<i
				 <<selerror("InvalidVectorElement", 1);

	VectorNode *vn = (VectorNode *)me->child(0);
	return (vn->*vn->eval)()[i];
} // VectorComponentArith()

OPCLASS2_A(VectorComponent)


static double ScalarProductArith(OpNode me)
{
	VectorNode *v1 = (VectorNode *)me->child(0),
			   *v2 = (VectorNode *)me->child(1);
	return (v1->*v1->eval)() * (v2->*v2->eval)();
} // ScalarProductArith()

OPCLASS2_A(ScalarProduct)


static double EvalLogicAsArith(OpNode me)
{ return me->logic(me) ? 1. : 0.; }

#define OPCLASS3_IMP(op)											\
op::op() { registerLogicFunc( op##Logic );							\
		   registerArithFunc(EvalLogicAsArith); }					\
op::op(const OpNode n1, const OpNode n2, const OpNode n3) {			\
	const OpNode nv[] = { n1, n2, n3 }; addChildren(3, nv);			\
registerLogicFunc( op##Logic );										\
registerArithFunc(EvalLogicAsArith); }								\
op::~op() {}

static bool ConeFilterLogic(OpNode me)
{
	VectorNode *axis = (VectorNode *)me->child(0),	//	axis
			   *pnt = (VectorNode *)me->child(2);	//	test vector
	double	   halfOpeningAngle = me->child(1)->arith(me->child(1));

	//Check that the current attitude is defined, if not, return false
	if (isnan(((pnt->*pnt->eval)()).x()) || isnan(((pnt->*pnt->eval)()).y()) ||
			isnan(((pnt->*pnt->eval)()).z()))
	{
		return false;
	}

	//Check if the inner product is not greater than one. If it is bigger than one (error propagation),
	//return the acos of 1.0
	if ((axis->*axis->eval)() * (pnt->*pnt->eval)() > 1.0 )
	{
		return acos(1.0) <= halfOpeningAngle;
	}

	return acos((axis->*axis->eval)() * (pnt->*pnt->eval)()) <=
				halfOpeningAngle;

} // ConeFilterLogic()

OPCLASS3_IMP(ConeFilter)

#define OPCLASS1_A(op)												\
op::op() { registerArithFunc( op##Arith ); }						\
op::op(const OpNode n1) { OpNode nv[] = { n1 }; addChildren(1, nv);	\
registerArithFunc( op##Arith ); }									\
op::~op() {}

static double VectorNormArith(OpNode me)
{
	VectorNode *v = (VectorNode *)me->child(0);

	return (v->*v->eval)().norm();
} // VectorNormArith()

OPCLASS1_A(VectorNorm)

