//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	DataFilter.cc:	--- implementation of DataFilter class
//
//
//	CHANGE HISTORY:
//		UL, 1998-01-06: 1.0.0: - first version
//		UL, 1998-04-06: 1.1.0: - bug fix in `doDestructiveFiltering'
//		UL, 1998-05-19: 1.2.0: - minor changes
//		UL, 1998-06-28: 1.3.0: - bug fix: contents of new flag column
//								 was not initialized to 0
//		UL, 1998-08-27: 1.4.0: - bug fix: clearing FLAG column resulted in
//								 crash when filtered table was empty
//		UL, 1999-03-03: 1.5.0: - changed enforced by Lowmem Dal
//		UL, 1999-04-16: 1.6.0: - changed Filter -> DataFilter to resolve
//								 name conflict with Cal
//		UL, 1999-10-08: 1.7.0: - expandSelectExpression: do not try
//								 to replace '#ROW'
//		UL, 1999-11-03: 1.8.0: - added support for new unsigned types
//							     in dal>=1.80
//		UL, 2000-07-03: 1.9.0: - several changes related to support of
//								 string columns
//		UL, 2000-10-30: 1.10.0:- ExprColumnConstructur did not understand
//								 attribute references in expressions
//		UL, 2002-07-19: 1.11.0:- minor C++ standard compliance fix w.r.t
//								 preprocessor usage
//
//	AUTHOR:
//		U. Lammers (UL), Tue Jan 06 11:48:12 MET DST 1998
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/service/DataFilter.cc,v 1.6 2009/06/26 15:51:59 aibarra Exp $
//
#include "Tracer.h"
#include "../selcore/basicdefs.h"
#include "DataFilter.h"
#include "ServiceUtils.h"

#if !defined(NO_SAS)
#include "SelectError.h"
#include<cstring>
SAS_DEBUG(#include <iostream>)

DataFilter::DataFilter() : _dalTableDescriptor(0), _evaluator(0) {}

//	construct filter with boolean expression
DataFilter::DataFilter(const std::string &expression)
		: _expression(expression), _dalTableDescriptor(0),
		  _evaluator(0)
{
} // DataFilter::DataFilter

std::string DataFilter::expandSelectionExpression(
											const std::string &expression,
											const Table &table)
{
	std::vector<std::string> special;
	special.push_back("ROW");

	return ServiceUtils::attributeFreeExpression(expression, &table, &special);
} // DataFilter::expandSelectionExpression()


static int bitPosition(Column *column, short flagBitPos)
{
	unsigned nDigits = 0;
	switch (column->dataType()) {
		case Dal::Int8  : nDigits = 8; break;
		case Dal::Int16 :
		case Dal::Uint16: nDigits = 16; break;
		case Dal::Int32 :
		case Dal::Uint32: nDigits = 32; break;
		default: selerrstr<<"Flag column must be of "
							"integral type"<<selerror("NotIntegralType");
				 SELSTDTHROW; break;
	}

	if (flagBitPos == -1 || static_cast<unsigned>(flagBitPos) > nDigits - 1)
		return nDigits - 1;

	return flagBitPos;
} // bitPosition()


template<class ColType>
static inline unsigned long whatMaskValue(ColType data, short flagBitPos)
{
	int ndigits = 8 * sizeof(data);
	return 1 << ((flagBitPos < 0 || flagBitPos > ndigits - 1)
				? ndigits - 1 : flagBitPos);
} // whatMaskValue()


template<class ColType>
static inline void selectRows(ColType *data, unsigned long from,
							  unsigned long nRows, unsigned long maskValue)
{
	unsigned maxTo = from + nRows;
	for (unsigned long n = from; n < maxTo; ++n)
		data[n] |= maskValue;
} // unselectRows()


template<class ColType>
static inline void unselectRows(ColType *data, unsigned long from,
								unsigned long nRows, unsigned long maskValue)
{
	unsigned maxTo = from + nRows;
	for (unsigned long n = from; n < maxTo; ++n)
		data[n] &= ~maskValue;
} // selectRows()


enum SelectModifier { SELECT, UNSELECT };
void setSelectStatus(Column *flagColumn, ColumnData *cd, unsigned long from,
					 unsigned long nRows, short flagBitPos,
					 SelectModifier selectModifier)
{
	#define CODE(dalType, type)												\
		case Dal:: dalType: {												\
			type *data = cd->type##Data();									\
			unsigned long maskValue = whatMaskValue(*data, flagBitPos);		\
			if (selectModifier == SELECT)									\
				selectRows(data, from, nRows, maskValue);					\
			else															\
				unselectRows(data, from, nRows, maskValue);					\
			break; }

	switch (flagColumn->dataType()) {
		CODE(Int8, int8)
		CODE(Int16, int16)
		CODE(Uint16, uint16)
		CODE(Int32, int32)
		CODE(Uint32, uint32)
		default: selerrstr<<"Flag column `"<<flagColumn->qualifiedName()
						  <<"' must be of integral type"
						  <<selerror("NotIntegralType");
				 SELSTDTHROW;
				 break;
	}

	#undef CODE
} // setSelectStatus()



unsigned long DataFilter::doFilter(Table &table) const
{
	std::vector<DataFilter::RowRange> rowRangeVec;

	//	do the selection
	unsigned long nsel = this->doSelect(table, rowRangeVec);

	//	first move blocks of selected events upwards
	unsigned long toRow = 0;
	unsigned nRows = rowRangeVec.size();
	SAS_DEBUG(cerr<<"Moving "<<nRows<<" row blocks ... "<<flush;)
	for (unsigned n=0; n<nRows; ++n) {
		unsigned long nRows = rowRangeVec[n].stopRow -
							  rowRangeVec[n].startRow + 1;
		table.copyRows(rowRangeVec[n].startRow, toRow, nRows);
		toRow += nRows;
	}
	SAS_DEBUG(cerr<<"done\n";)

	//	delete block of non-selected rows at the end
	SAS_DEBUG(cerr<<"Deleting block of unselected rows at the end ... "<<flush;)
	table.deleteRows(toRow, table.rows() - toRow);
	SAS_DEBUG(cerr<<"done\n";)

	//	check if flag column exists - if yes and it was created by
	//	this task, delete it
	if (table.hasAttribute(DataFilter::FLAGCOLUMN)) {
		//	table contains flag column
		std::string flagColumnName =
					  table.attribute(DataFilter::FLAGCOLUMN)->asString();
		if (table.hasColumn(flagColumnName) &&
								table.hasAttribute(DataFilter::FLAGCREATE))
		{
			if (table.attribute(DataFilter::FLAGCREATE)->asBool())
				//	table has flag column which was created in a prious
				//	filter run - delete it
				table.deleteColumn(table.columnNumber(flagColumnName));
			else
				if (table.hasAttribute(DataFilter::FLAGBIT)) {
					short flagBitPos =
						  table.attribute(DataFilter::FLAGBIT)->asInt();
					//	table has flag column but this was not created in
					//	a prious run - mark all events as selected
					ColumnData *cd = table.column(flagColumnName)->data();
					setSelectStatus(table.column(flagColumnName), cd,
									0, table.rows(), flagBitPos,
									SELECT);
					delete cd;
				} else
					selerrstr<<"Table `"<<table.qualifiedName()<<"' has "
							   "been non-destructively filtered but "
							   "cannot figure which bit position was "
							   "used to flag selected rows - attribute `"
							 <<DataFilter::FLAGBIT<<"' does not exist"
							 <<selwarning("AttributeMissing");
		}
	}

	//
	//	delete attribute
	//
	#define DELATT(a)	if (table.hasAttribute(a)) table.deleteAttribute(a)

	DELATT(DataFilter::FLAGCOLUMN);
	DELATT(DataFilter::FLAGBIT);
	DELATT(DataFilter::FLAGCREATE);

	#undef DELATT

	return nsel;
} // DataFilter::doFilter()


//template<class Type>
//static void setAttribute(const Table &table, const std::string &name,
//						 const std::string &desc, const Type &value)
#define setAttribute(table, name, desc, value)				\
	{ Attribute *att = table.hasAttribute(name)				\
						? table.attribute(name)				\
						: table.addAttribute(name, desc);	\
	*att = value; }


const std::string DataFilter::DF_FLAG_COLUMN_NAME("FLAG");
const std::string DataFilter::FLAGCOLUMN("FLAGCOL");
const std::string DataFilter::FLAGBIT("FLAGBIT");
const std::string DataFilter::FLAGCREATE("FLAGCRT");

unsigned long DataFilter::doSelect(Table &table,
					   const std::string &flagColumnName,
					   short flagBitPos) const
{
	std::vector<DataFilter::RowRange> rowRangeVec;

	//	do the selection
	unsigned long nsel = this->doSelect(table, rowRangeVec);

	Column *flagColumn;
	if (!table.hasColumn(flagColumnName)) {
		//	column does not exist, add it
		std::vector<unsigned long> s; s.push_back(1);
		flagColumn = table.addColumn(flagColumnName, Dal::Int8,
						 "event status: 0/1 - selected/non-selected", "", s);
		//	initialize column to 0
		ColumnData *cd = flagColumn->data();
		int8 *flagVal = cd->int8Data();
		unsigned nRows = table.rows();
		for (unsigned n=0; n<nRows; flagVal[n++] = 0) ;
		delete cd;

		setAttribute(table, FLAGCREATE,
					 "FLAG column created by DataFilter?", true);
	} else {
		flagColumn = table.column(flagColumnName);
		setAttribute(table, FLAGCREATE,
					 "FLAG column created by DataFilter?", false);
	}

	//
	//	set the flag bit thereby marking all events as non-selected
	//
	ColumnData *cd = flagColumn->data();
	setSelectStatus(flagColumn, cd, 0, table.rows(), flagBitPos, UNSELECT);

	//	mark selected events by clearing bit in flag column
	for (unsigned n=0; n<rowRangeVec.size(); ++n)
		setSelectStatus(flagColumn, cd, rowRangeVec[n].startRow,
						rowRangeVec[n].stopRow - rowRangeVec[n].startRow + 1,
						flagBitPos, SELECT);

	delete cd;

	//
	//	add attributes to hold the name of flag column and bit position
	//
	setAttribute(table, FLAGCOLUMN, "flag column name", flagColumnName);
	setAttribute(table, FLAGBIT, "flag bit position",
				 bitPosition(flagColumn, flagBitPos));

	return nsel;
} // DataFilter::doSelect()


//
//	filter a table
//
unsigned long DataFilter::doSelect(const Table &table,
								   std::vector<RowRange> &rowRangeVec) const
{
	double						**valPtr;
	void						**dataVec;
	ServiceUtils::ConvFuncPtr	*convFuncVec;
	unsigned					nNumericIds;

	char						**textValPtr;
	unsigned					nTextIds;
	char						**textDataVec;
	unsigned					*textFieldLength;

	//	retarted instantiation of Evaluator object
	delete _evaluator;
	delete _dalTableDescriptor;
	_dalTableDescriptor = new DalTableDescriptor(&table);
	_expression = DataFilter::expandSelectionExpression(_expression, table);
	_evaluator = new Evaluator(_expression, _dalTableDescriptor);

	std::vector<ColumnData *> cdv =
		ServiceUtils::prepareEvaluation(_evaluator, table, valPtr, nNumericIds,
										dataVec, convFuncVec,
										textValPtr, nTextIds, textDataVec,
										textFieldLength);
	bool	 startNewRowRange = true;
	RowRange rowRange;
	unsigned long nsel = 0;

	//
	//	--------------------( the event loop )----------------------------
	//
	unsigned nRows=table.rows();
	for (unsigned n=0; n<nRows; ++n) {
		//	assign numeric variable values
		for (unsigned i=0; i<nNumericIds; ++i)
			*valPtr[i] = (*convFuncVec[i])(dataVec[i], n);

		//	assign row number
		*valPtr[nNumericIds] = DBL(n+1);

		//	assign string variable values
		for (unsigned i=0; i<nTextIds; ++i)
			std::strncpy(textValPtr[i], textDataVec[i]+n*textFieldLength[i],
						 textFieldLength[i]);

		//	evaluate expression
		if (_evaluator->evalLogic()) {
			//	selected
			if (startNewRowRange) {
				startNewRowRange = false;
				rowRange.startRow = n;
			}
		} else
			//	not selected
			if (!startNewRowRange) {
				startNewRowRange = true;
				rowRange.stopRow = n-1;
				nsel += rowRange.stopRow - rowRange.startRow + 1;
				rowRangeVec.push_back(rowRange);
			}
	}
	if (!startNewRowRange) {
		rowRange.stopRow = table.rows()-1;
		nsel += rowRange.stopRow - rowRange.startRow + 1;
		rowRangeVec.push_back(rowRange);
	}

	//
	//	be nice and free-up memory
	//
	delete [] valPtr;
	delete [] dataVec;
	delete [] convFuncVec;
	delete [] textValPtr;
	delete [] textDataVec;
	delete [] textFieldLength;
	unsigned ncdv = cdv.size();
	for (unsigned n=0; n<ncdv; delete cdv[n++]) ;

	return nsel;
} // DataFilter::doSelect()


DataFilter::~DataFilter()
{
	delete _evaluator;
	delete _dalTableDescriptor;
} // DataFilter::~DataFilter()

#endif
