//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	ServiceF90cxxGlue.cc --- implementation Fortran90/C++ interface
//
//	DESCRIPTION:
//		This is the implementation of the filter/arrayEval class interface.
//
//	CHANGE HISTORY:
//		UL, 1998-03-25: 1.0.0: - skeleton
//		UL, 1998-05-28: 1.1.0: - added ArrayEvaluator
//		UL, 2002-07-09: 1.2.0: - added ExprColumnGenerator
//
//	AUTHOR:
//		U. Lammers, Wed Mar 25 11:15:00 MET 1998
//
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/service/ServiceF90cxxGlue.cc,v 1.5 2002/12/17 17:48:11 jbakker Exp $
//

#include <cassert>
#include "stringutil.h"
#include "DataFilter.h"
#include "SelectError.h"
#include "ServiceF90cxxGlue.h"

//
//	---------------------------( DataFilter )--------------------------------
//

DataFilter *filter = 0;	 // a single global instance of filter object

void CXXFUNC(initfilter, initFilter)
			(const char *expression, Integer3 expressionLength)
{
	delete filter;
	filter = new DataFilter(trim(std::string(expression, expressionLength)));
	assert(filter);
} // initFilter()


Integer3 CXXFUNC(dofilter, doFilter)(Table *const *table)
{
	if (!filter) {
		selerrstr<<"No filter object has been "
				   "initialized - please call first `initFilter(<expression>)'"
				 <<selerror("ImproperUsage");
		SELSTDTHROW;
	}

	//	do the selection
	return filter->doFilter(**table);
} // doFilter()


Integer3 CXXFUNC(doselectmodify, doSelectModify)
									(Table *const *table,
		 							 const char *flagColumnName,
									 const Integer2 *flagBitPos,
									 Integer3 flagColumnLength)
{
	std::string flagColName	= flagColumnName
							? trim(std::string(flagColumnName, flagColumnLength))
							: DataFilter::DF_FLAG_COLUMN_NAME;
	short bitPos		= flagBitPos
							? static_cast<unsigned short>(*flagBitPos)
							: DataFilter::DF_FLAG_BITPOS;

	// do the selection
	return filter->doSelect(**table, flagColName, bitPos);
} // doSelectModify()


Integer3 CXXFUNC(doselectconst, doSelectConst)
			(const Table *const *table,
			 Dope<Integer3, 1> *fromRow, Dope<Integer3, 1> *toRow)
{
	if (!filter) {
		selerrstr<<"No filter object has been "
				   "initialized - please call first `initFilter(<expression>)'"
				 <<selerror("ImproperUsage");
		SELSTDTHROW;
	}

	std::vector<DataFilter::RowRange> rowRangeVec;

	//	do the selection
	unsigned long nsel = filter->doSelect(**table, rowRangeVec);

	unsigned nBlocks = rowRangeVec.size();
	Integer3 *fromRowVec = new Integer3[nBlocks],
			 *toRowVec = new Integer3[nBlocks];

	for (unsigned n=0; n<nBlocks; ++n) {
		fromRowVec[n] = rowRangeVec[n].startRow;
		toRowVec[n] = rowRangeVec[n].stopRow;
	}

	std::vector<unsigned long> extent; extent.push_back(nBlocks);
	*fromRow = Dope<Integer3, 1>(fromRowVec, extent);
	*toRow = Dope<Integer3, 1>(toRowVec, extent);

	return nsel;
} // doFilterConstTable()

void CXXFUNC(selectreleasememory, selectReleaseMemory)
												(Dope<Integer3, 1> *fromRow,
												 Dope<Integer3, 1> *toRow)
{
	delete [] fromRow->addr; fromRow->addr = 0;
	delete [] toRow->addr; toRow->addr = 0;
} // selectReleaseMemory()


#include "ArrayEvaluator.h"
//
//	---------------------------( ArrayEvaluator )----------------------------
//
ArrayEvaluator *arrayEvaluator = 0; // single globale instance of object

void CXXFUNC(initarrayeval, initArrayEval)
			(const char *expression, Integer3 expressionLength)
{
	delete arrayEvaluator;
	arrayEvaluator = new ArrayEvaluator(trim(std::string(expression,
										expressionLength)));
	assert(arrayEvaluator);
} // initArrayEval()


void CXXFUNC(doarrayeval, doArrayEval)
			(Dope<Array *, 1> *inArray, Integer3 *lenInArray, Array **outArray)
{
	if (!arrayEvaluator) {
		selerrstr<<"No ArrayEvaluator object has been "
				   "initialized - please call first "
				   "`initArrayEval(<expression>)'"
				 <<selerror("ImproperUsage");
		SELSTDTHROW;
	}

	std::vector<Array *> arrayVec(*lenInArray);
	for (unsigned n=0; n<static_cast<unsigned>(*lenInArray); ++n)
		arrayVec[n] = (inArray->addr)[n];

	//	do the job
	arrayEvaluator->doIt(arrayVec, **outArray);
} // doArrayEval()


#include "ExprColumnGenerator.h"
//
//	-----------------------( ExprColumnGenerator )----------------------------
//
static ColumnGenerator *exprColGen = 0; // single globale instance of object

void CXXFUNC(initexprcolgen, initExprColGen)
			(const char *expression, const char *colName,
			 const Integer3 *colType,
			 Integer3 expressionLength, Integer3 colNameLength)
{

	delete exprColGen;
	std::string expr = trim(std::string(expression, expressionLength)),
		   cname = colName ? trim(std::string(colName, colNameLength)) : "TEMP";
	Dal::DataType ctype = colType ? static_cast<Dal::DataType>(*colType)
								  : Dal::Real64;

	exprColGen = new ExprColumnGenerator(expr, cname, ctype);
	assert(exprColGen);
} // initExprColGen()


void CXXFUNC(doexprcolgen, doExprColGen)
			(const Table **inTab, Table **outTab)
{
	if (!exprColGen) {
		selerrstr<<"No ExprColumnGenerator object has been "
				   "initialized - please call first "
				   "`initExprColGen(<expression>, <colName>, <colType>)'"
				 <<selerror("ImproperUsage");
		SELSTDTHROW;
	}

	//	do it
	exprColGen->doIt(**inTab, **outTab);
} // doExprColGen()
