!------------------------------------------------------------------------------
!
!                       XMM Science Analysis System (SAS)
!                     (c) 1997-2010 European Space Agency
!
!------------------------------------------------------------------------------
!
!   exprcolgen.f90 --- Generate new columns in a table
!
!   DESCRIPTION:
!       A module with functionality to generate new columns in a table
!       driven by arithmetic expressions involving other column names,
!       operators, functions, etc.
!
!   EXPORTED FUNCTIONS:
!       initExprColGen --- initialize the column generator with an expression
!       doExprColGen   --- perform the operation
!
!   CHANGE HISTORY:
!       UL, 2002-07-09: 1.0.0: - first version
!
!   AUTHOR:
!       U. Lammers, Tue Jul  9 11:52:05 MEST 2002
!
!   $Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/service/exprcolgen.f90,v 1.1 2002/07/09 14:06:32 ulammers Exp $
!
module ExprColGen
    implicit none

    interface
        !
        !   SYNOPSIS:
        !      initExprColGen(expression, colName, colType)
        !
        !   PURPOSE:
        !      Initialize the column generator object with an arithmetic
        !      expression plus the name and type of the column to be
        !      newly constructed.
        !
        !   ARGUMENTS:
        !      expression: a string which must represent a valid arithmetic
        !                  expression; an error condition is raised if it
        !                  does not; the expression must adhere to the general
        !                  selectlib syntax rules for arithmetic and logical
        !                  expressions detailed in the package documentation.
        !                  The expression may contain names of existing
        !                  columns in the table passed as argument <inTable>
        !                  to below subroutine 'doExprColGen'.
        !                  An example of a valid expression involving the
        !                  existing columns 'RAWX' and 'RAWY' in the input
        !                  table is:
        !                      sqrt((RAWX-1)**2 + (RAWY-1)**2)
        !      colName   : the name of the new column to be constructed
        !                  (optional - d/f: "TEMP")
        !      colType   : the type of the new column to be constructed
        !                  (optional - d/f: DOUBLE)
        !
        subroutine initExprColGen(expression, colName, colType)
             character(len=*), intent(in)               :: expression
             character(len=*), intent(in), optional     :: colName
             integer, intent(in), optional              :: colType
        end subroutine


        !
        !   SYNOPSIS:
        !      doExprColGen(inTable, outTable)
        !
        !   PURPOSE:
		!      Create a new column in a table potentially using data
		!      from other columns in another table; routine 'initExprColGen'
        !      must have been called prior to 'doExprColGen'; an error
        !      condition shall be raised otherwise
        !
        !   ARGUMENTS:
        !      inTable  : reference to exisiting table containing the columns
        !                 referred to in the arithmetic expression passed
        !                 to routine 'initExprColGen'.
        !      outTable : the table to which the newly created column <colName>
        !                 will be added; <outTable> and <inTable> can be
        !                 identical.
        !
        subroutine doExprColGen(inTable, outTable)
            use types
            use dal
            type(TableT), intent(in)                   :: inTable, outTable
        end subroutine
    end interface

end module ExprColGen






