//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	ASCregion.h: --- read/write ASC region files
//
//	DESCRIPTION:
//		Instances of class ASCregionTable represent region filters as
//		specified in "ASC FITS File Designers' Guide", J. McDowell and
//		A. Rots.
//
//	CHANGE HISTORY:
//		UL, 25/03/98: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers (UL), Wed Mar 25 16:52:23 MET 1998
//
//	REFERENCES:
//		[1]: FITS Region Binary Table Design, ASC-FITS-REGION-1.0,
//			 J. McDowell, A. Rots
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/utils/ASCregion.h,v 1.3 2002/01/29 16:18:40 ulammers Exp $
//

#ifndef ASCREGIONFILE_H
#define ASCREGIONFILE_H

#include <string>
#include <vector>
#include <map>
#include "Dal.h"
#include "Region.h"

class ASCregionTable : public virtual ShapeBasics {
public:
	ASCregionTable() {}
	ASCregionTable(const std::string &setName, const std::string &xCoordName,
				   const std::string &yCoordName, bool closeSet = true);
	virtual ~ASCregionTable() {}

	enum DataType { SHAPE = 0, X, Y, R, ROTANG, COMPONENT, INCLUDE };
	std::string setName() const { return _setName; }
	std::string xcoord() const { return _xCoord; }
	std::string ycoord() const { return _yCoord; }

protected:
	//	data members
	DataSet		*_set;
	Table		*_table;
	std::string	_setName, _xCoord, _yCoord;
	bool		_closeSet;		//	close set in destructor?
}; // ASCRegionTable


class InASCregionTable : public virtual ASCregionTable {
public:
	InASCregionTable() {}

	//	construct from data set name - entails opening of it
	InASCregionTable(const std::string &setName);

	//	construct from open data set
	InASCregionTable(DataSet *dataSet);

	//	construct from table
	InASCregionTable(Table *table);

	//	destructor
	~InASCregionTable();

	//	how many elements are there?
	unsigned elements() const;

	//	how many components are there?	
	unsigned components() const;

	//	to which component belongs an element?
	unsigned component(unsigned elemenNr);

	//	is an element included or excluded?
	bool included(unsigned elementNr) const;

	//	what shape has an element?
	ShapeBasics::ShapeType whatShape(unsigned elementNr) const;

	//	get data for an element
	std::vector<float> data(DataType dataType, unsigned elementNr) const;

private:
	std::vector<Column *>		_column;	 //	to be indexed with DataType
	std::vector<std::string>	_columnNames;//	to be indexed with ShapeType
	typedef std::map<unsigned, unsigned, std::less<unsigned> >	MapType;
	MapType				_compMap;

	void initFromTable();
}; // InASCregionTable


class OutASCregionTable : public virtual ASCregionTable {
public:
	OutASCregionTable() {}

	//	construct from set name - entails opening a new data set
	OutASCregionTable(const std::string &setName,
					  const std::string &xCoordName = "X",
					  const std::string &yCoordName = "Y");

	//	construct from open set
	OutASCregionTable(DataSet *dataSet, const std::string &xCoordName = "X",
					  const std::string &yCoordName = "Y");

	//	construct from table
	OutASCregionTable(Table *table, const std::string &xCoordName = "X",
					  const std::string &yCoordName = "Y");

	//	destructor
	~OutASCregionTable();

	friend OutASCregionTable &operator<<(OutASCregionTable &,
										 const GeoRegion &);
}; // OutASCregionTable

//
//	PURPOSE:
//		Read contents of CXC region file [1] and initialize passed
//		GeoRegion object with it; Shapes are dynamically created; ownership
//		is transfered to caller who is responsible for their deletion
//	
InASCregionTable &operator>>(InASCregionTable &, GeoRegion &);

//
//	PURPOSE:
//		Write contents of GeoRegion object to a CXC region file [1]
//
OutASCregionTable &operator<<(OutASCregionTable &, const GeoRegion &);

#endif



