//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	Region.cc:	--- implementation of region class
//
//	DESCRIPTION:
//		Implementation of region class
//
//	CHANGE HISTORY:
//		UL, 1998-03-25: 1.0.0: - first version
//		UL, 2000-05-09: 1.1.0: - toSelectExpression() is trying to find
//								 out number of significant digits
//		UL, 2000-10-24: 1.2.0: - Selecting::toSelectExpression() throws
//								 now exceptions (this fixes SOC-SPR SPR 1708:
//								 Seg fault when importing an invalid 1-d
//								 region from grace)
//		UL, 2000-11-13: 1.3.0: - added Annulus<float>::toSelectionExpression()
//		UL, 2001-02-05: 1.4.0: - added Shape::contains()
//		UL, 2002-01-28: 1.5.0: - added EllipticalAnnulus::toSelectExpression()
//								 (override base class method)
//		UL, 2002-07-11: 1.6.0: - added
//									Component::toSelectExpression()
//									Component::contains()
//									GeoRegion::toSelectExpression()
//									GeoRegion::contains()
//							   - implemented funtools/CIAO complaint region
//								 combination policy
//
//	AUTHOR:
//		U. Lammers (UL), Wed Mar 25 16:52:23 MET 1998
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/utils/Region.cc,v 1.10 2014/04/15 15:01:34 aibarra Exp $
//

#include <algorithm>	// max()
#include <functional>

#include "boost/functional.hpp"	//	stdlib not good enough
#include "Region.h"
#include "errstr.h"
#include "stringutil.h"
#include "common.h"

#include <cassert>
/* AI gcc porting
 * fabs was not included.
 */
#include <cmath>

using namespace SelectError;
using namespace ShapeUtils;
using namespace std;

//
//	------------------------( ShapeBasics )------------------------------------
//
string ShapeBasics::toString(ShapeBasics::ShapeColor color)
{
	static const char* colorString[] = { "black", "white", "red", "green",
										 "blue", "cyan", "magenta", "yellow" };

	assert(UNT(color) < NUM_OF(colorString));

	return colorString[UNT(color)];
} // ShapeBasics::toString()


ShapeBasics::ShapeColor ShapeBasics::toColor(const string &colorString)
{
    std::string col(lowercase(colorString));
    if (col == "black") return ShapeBasics::BLACK;
    else if (col == "red") return ShapeBasics::RED;
    else if (col == "green") return ShapeBasics::GREEN;
    else if (col == "blue") return ShapeBasics::BLUE;
    else if (col == "cyan") return ShapeBasics::CYAN;
    else if (col == "magenta") return ShapeBasics::MAGENTA;
    else if (col == "yellow") return ShapeBasics::YELLOW;
    else if (col == "white") return ShapeBasics::WHITE;
	else {
		selerrstr<<"Unknown color `"<<colorString<<"'"
				 <<selerror("UnknownColor");
		SELSTDTHROW;
	}

    return ShapeBasics::BLACK;
} // ShapeBasics::toColor()


//
//	------------------------( Selecting )-------------------------------------
//
std::string Selecting::toSelectExpression(const std::string &axisName, double, double)
const throw(UnexpressionableRegion)
{
	selerrstr<<"Region cannot be converted into a 1-D selection "
			   "expression on column `"
			 <<axisName<<"'"<<selerror("CannotConstruct1DExpression", 2);
	SELTHROW(UnexpressionableRegion);

	return "";
} // Selecting::toSelectExpression()


std::string Selecting::toSelectExpression(const std::string &xAxisName,
										  const std::string &yAxisName,
										  double, double, double, double) const
throw(UnexpressionableRegion)
{
	selerrstr<<"Region cannot be converted into 2-D selection expression on "
			   " columns `"<<xAxisName<<"/"<<yAxisName<<"'"
			 <<selerror("CannotConstruct2DExpression", 2);
	SELTHROW(UnexpressionableRegion);

	return "";
} // Selecting::toSelectExpression()


std::string Selecting::prefix(const std::string &axisName, bool included) const
{
	return (included ? "(" : "!(") + axisName + ") IN ";
} // Selecting::prefix()


std::string Selecting::prefix(const std::string &xAxisName,
							  const std::string &yAxisName,
							  bool included) const
{
	return (included ? "(" : "!(") + xAxisName + ',' + yAxisName + ") IN ";
} // Selecting::prefix()


std::string Selecting::convertPosition(double x, double binX, double minX,
									   double y, double binY, double minY)
const
{
	return ::toString((x-1.) * binX + minX) + ',' +
		   ::toString((y-1.) * binY + minY);
} // Selecting::convertPosition()


std::string Selecting::convertDimension(double dim, double f) const
{
	return ::toString(dim * f);
} // Selecting::convertDimension()


//
//	----------------------------( Shapes )-------------------------------------
//
/* AI gcc4 porting
 * In all the toSelectExpression methods, we had to remove explicetily the
 * type parameter and change it for the Template T...
 */
template<class T>
std::string Shape<T>::toSelectExpression(const std::string &axisName,
											 double bin,
											 double min) const
			throw(UnexpressionableRegion)
{
	return Selecting::toSelectExpression(axisName, bin, min);
} // Shape<float>::toSelectExpression()

template std::string Shape<float>::toSelectExpression(const std::string &axisName,
											 double bin,
											 double min) const
			throw(UnexpressionableRegion);
template std::string Shape<double>::toSelectExpression(const std::string &axisName,
											 double bin,
											 double min) const
			throw(UnexpressionableRegion);
/*
std::string Shape<double>::toSelectExpression(const std::string &axisName,
											  double bin, double min) const
	throw(UnexpressionableRegion)
{
	return Selecting::toSelectExpression(axisName, bin, min);
} // Shape<float>::toSelectExpression()
*/

template<class T>
std::string Shape<T>::toSelectExpression(const std::string &xAxisName,
											 const std::string &yAxisName,
											 double binX, double binY,
											 double minX, double minY) const
	throw(UnexpressionableRegion)
{
	return prefix(xAxisName, yAxisName, included()) + "point(" +
			convertPosition(_refPoint.x(), binX, minX,
							_refPoint.y(), binY, minY) + ')';
} // Shape<float>::toSelectExpression()
template std::string Shape<float>::toSelectExpression(const std::string &xAxisName,
											 const std::string &yAxisName,
											 double binX, double binY,
											 double minX, double minY) const
	throw(UnexpressionableRegion);
template std::string Shape<double>::toSelectExpression(const std::string &xAxisName,
											 const std::string &yAxisName,
											 double binX, double binY,
											 double minX, double minY) const
	throw(UnexpressionableRegion);


/*
std::string Shape<double>::toSelectExpression(const std::string &xAxisName,
											  const std::string &yAxisName,
											  double binX, double binY,
											  double minX, double minY) const
	throw(UnexpressionableRegion)
{
	return prefix(xAxisName, yAxisName, included()) + "point(" +
			convertPosition(_refPoint.x(), binX, minX,
							_refPoint.y(), binY, minY) + ')';
} // Shape<double>::toSelectExpression()
*/
template<class T>
std::string Line<T>::toSelectExpression(const std::string &axisName,
									    double bin, double min) const
	throw(UnexpressionableRegion)
{
	if (this->refPoint().y() == _p2.y()) {
		//	a simple interval
		T x0 = this->refPoint().x(),
		  x1 = _p2.x();
		if (x0 > x1) {
			x0 = x1;
			x1 = this->refPoint().x();
		}
		unsigned digits = bin != 1.|| min != 0. ? 12 : 6;
		return Selecting::prefix(axisName, this->included())
				+ '[' + ::toString(x0*bin+min, digits) + ':' +
						::toString(x1*bin+min, digits) + ']';
	}

	return axisName +
			(this->refPoint().y() < _p2.y() ? "<="	//	LEFT_OF_LINE
									  : ">=")	// RIGHT_OF_LINE
			+ ::toString(.5*bin*(this->refPoint().x() + _p2.x())+min, 6);
} // Line<T>::toSelectExpression()


template std::string Line<float>::toSelectExpression(
												const std::string &axisName,
											    double bin, double min) const
	throw(UnexpressionableRegion);
template std::string Line<double>::toSelectExpression(
												const std::string &axisName,
											    double bin, double min) const
	throw(UnexpressionableRegion);

//
//	--------------------------( Component )------------------------------------
//
bool Component::contains(const FPoint &p) const
{
	//	point is contained if contained in any of the constituent shapes;
	//	boost version of bind2nd/mem_fun are required because stdlib
	//	gives reference to reference problem (see
	//	http://www.boost.org/libs/functional/binders.html
	//	for details)
	bool incTest = find_if(begin(), end(),
					   boost::bind2nd(boost::mem_fun(&FShape::contains), p))
							!= end();

	return included() ? incTest : !incTest;
} // Component::contains()


std::string Component::toSelectExpression(const std::string &xAxis,
										  const std::string &yAxis,
										  double binX, double binY,
										  double minX, double minY) const
												throw(UnexpressionableRegion)
{
	string expr;
	unsigned N = size(), n = 0;
	for (; n<N; ++n) {
		if (n)
			expr += "||";
		expr += (*this)[n]->toSelectExpression(xAxis, yAxis, binX, binY,
											   minX, minY);
	}

	if (n > 1)
		// enclose in parentheses
		expr = '(' + expr + ')';

	//	negate for 'excluded' regions'
	if (N && !included())
		expr = '!' + expr;

	return expr;
} // Component::toSelectExpression()


//
//	------------------------( GeoRegion )-------------------------------------
//
bool GeoRegion::contains(const FPoint &p) const
{
	//	point is contained if contained in all components
	bool incTest =
		UNT(count_if(begin(), end(),
				boost::bind2nd(boost::mem_fun_ref(&Component::contains), p)))
			== size();

	return included() ? incTest : !incTest;
} // GeoRegion::contains()


std::string GeoRegion::toSelectExpression(const std::string &xAxis,
										  const std::string &yAxis,
										  double binX, double binY,
										  double minX, double minY) const
												throw(UnexpressionableRegion)
{
	string expr;
	unsigned nc = 0;
	for (vector<Component>::const_iterator i=begin(); i!=end(); ++i) {
		string e = i->toSelectExpression(xAxis, yAxis, binX, binY, minX, minY);
		if (!e.empty()) {
			if (nc++)
				expr += "&&";
			expr += e;
		}
	}

	if (nc>=2)
		//	enclose in parentheses
		expr = '(' + expr + ')';

	//	negate for 'excluded' regions'
	if (nc && !included())
		expr = '!' + expr;

	return expr;
} // GeoRegion::toSelectExpression()


unsigned GeoRegion::maxPoints() const {
	unsigned nPoints = 0;
	for (unsigned comp=0; comp<this->size(); ++comp)
		for (unsigned element=0; element<(*this)[comp].size(); ++element) {
			std::vector<float> p = (*this)[comp][element]->pointsX();
			nPoints = max(nPoints, static_cast<unsigned>(p.size()));
		}

	return nPoints;
} // GeoRegion::maxPoints


bool GeoRegion::hasPolygon() const
{
	for (unsigned comp=0; comp<this->size(); ++comp)
		for (unsigned element=0; element<(*this)[comp].size(); ++element)
			if ((*this)[comp][element]->type() == ShapeBasics::POLYGON)
				return true;

	return false;
} // GeoRegion::hasPolygon()


unsigned GeoRegion::maxRadii() const {
	unsigned nRadii = 0;
	for (unsigned comp=0; comp<this->size(); ++comp)
		for (unsigned element=0; element<(*this)[comp].size(); ++element) {
			std::vector<float> r = (*this)[comp][element]->radii();
			nRadii = max(nRadii, static_cast<unsigned>(r.size()));
		}

	return nRadii;
} // GeoRegion::maxRadii


unsigned GeoRegion::maxRotAngles() const {
	unsigned nRotAngles = 0;
	for (unsigned comp=0; comp<this->size(); ++comp)
		for (unsigned element=0; element<(*this)[comp].size(); ++element) {
			std::vector<double> a = (*this)[comp][element]->rotAngles();
			nRotAngles = max(nRotAngles, static_cast<unsigned>(a.size()));
		}

	return nRotAngles;
} // GeoRegion::nRotAngles


unsigned GeoRegion::elements() const {
	unsigned nElements = 0;
	for (unsigned comp=0; comp<this->size();
			nElements += (*this)[comp++].size()) ;

	return nElements;
} // GeoRegion::elements


void GeoRegion::free() {
	//	tidy up
	for (unsigned n=0; n<size(); ++n) {
		unsigned M = (*this)[n].size();
		for (unsigned m=0; m<M; ++m) {
			delete (*this)[n][m];
			(*this)[n][m] = 0;
		}
	}
} // GeoRegion::free()


GeoRegion::~GeoRegion()
{
} // GeoRegion::~GeoRegion()

template<class T>
std::string Circle<T>::toSelectExpression(const std::string &xAxisName,
											  const std::string &yAxisName,
											  double binX,
											  double binY,
											  double minX,
											  double minY) const
	throw(UnexpressionableRegion)
{
	if (binX == binY)
		return Selecting::prefix(xAxisName, yAxisName, this->included()) + "circle(" +
				Selecting::convertPosition(this->_refPoint.x(), binX, minX,
							   this->_refPoint.y(), binY, minY) + ',' +
			   Selecting::convertDimension(_radius, binX) + ')';

	//	for different bin sizes, a circle becomes an ellipse
	return Selecting::prefix(xAxisName, yAxisName, this->included()) + "ellipse(" +
				Selecting::convertPosition(this->_refPoint.x(), binX, minX,
							   this->_refPoint.y(), binY, minY) + ',' +
			   Selecting::convertDimension(_radius, binX) + ',' +
			   Selecting::convertDimension(_radius, binY) + ",0)";
} // Circle<float>::toSelectExpression()

template std::string Circle<float>::toSelectExpression(const std::string &xAxisName,
											  const std::string &yAxisName,
											  double binX,
											  double binY,
											  double minX,
											  double minY) const
	throw(UnexpressionableRegion);

template<class T>
std::string Annulus<T>::toSelectExpression(const std::string &xAxisName,
											   const std::string &yAxisName,
											   double binX,
											   double binY,
											   double minX,
											   double minY) const
	throw(UnexpressionableRegion)
{
	if (binX == binY)
		return Selecting::prefix(xAxisName, yAxisName, this->included()) + "annulus(" +
				Selecting::convertPosition(this->_refPoint.x(), binX, minX,
							   this->_refPoint.y(), binY, minY) + ',' +
							   Selecting::convertDimension(_innerCircle.radius(), binX) + ',' +
							   Selecting::convertDimension(_outerCircle.radius(), binX) + ')';

	//	for different bin sizes, a Annulus becomes an ElliptAnnulus
	return Selecting::prefix(xAxisName, yAxisName, this->included()) + "elliptannulus(" +
			Selecting::convertPosition(this->_refPoint.x(), binX, minX,
							   this->_refPoint.y(), binY, minY) + ',' +
							   Selecting::convertDimension(_innerCircle.radius(), binX) + ',' +
							   Selecting::convertDimension(_innerCircle.radius(), binY) + ',' +
							   Selecting::convertDimension(_outerCircle.radius(), binX) + ',' +
							   Selecting::convertDimension(_outerCircle.radius(), binY) + ",0,0)";
} // Annulus<float>::toSelectExpression()

template std::string Annulus<float>::toSelectExpression(const std::string &xAxisName,
											   const std::string &yAxisName,
											   double binX,
											   double binY,
											   double minX,
											   double minY) const
	throw(UnexpressionableRegion);

template<class T>
std::string Ellipse<T>::toSelectExpression(const std::string &xAxisName,
											   const std::string &yAxisName,
											   double binX,
											   double binY,
											   double minX,
											   double minY) const
	throw(UnexpressionableRegion)
{
	return Selecting::prefix(xAxisName, yAxisName, this->included()) + "ellipse(" +
			Selecting::convertPosition(this->_refPoint.x(), binX, minX,
						   this->_refPoint.y(), binY, minY) + ',' +
						   Selecting::convertDimension(_radiusMajor, binX) + ',' +
						   Selecting::convertDimension(_radiusMinor, binY) + ',' +
						   ::toString(this->rotAngle()) + ')';
} // Ellipse<float>::toSelectExpression()

template std::string Ellipse<float>::toSelectExpression(const std::string &xAxisName,
											   const std::string &yAxisName,
											   double binX,
											   double binY,
											   double minX,
											   double minY) const
	throw(UnexpressionableRegion);

#if defined __GNUC__ && __GNUC__ < 4
std::string ElliptAnnulus<float>::toSelectExpression(
						 const std::string &xAxisName,
						 const std::string &yAxisName,
						 double binX, double binY,
						 double minX,
						 double minY) const
  throw(UnexpressionableRegion)
#else
template std::string ElliptAnnulus<float>::toSelectExpression(
							      const std::string &xAxisName,
							      const std::string &yAxisName,
							      double binX, double binY,
							      double minX,
							      double minY) const
  throw(UnexpressionableRegion);



template<class T>
std::string ElliptAnnulus<T>::toSelectExpression(
						 const std::string &xAxisName,
						 const std::string &yAxisName,
						 double binX, double binY,
						 double minX,
						 double minY) const
  throw(UnexpressionableRegion)
#endif
{
  return Selecting::prefix(
		xAxisName,
		yAxisName,
		this->included()) + "elliptannulus(" +
		Selecting::convertPosition(this->_refPoint.x(), binX, minX,
		    this->_refPoint.y(), binY, minY) + ',' +
		    Selecting::convertDimension(_innerEllipse.radiusMajor(), binX) + ',' +
		    Selecting::convertDimension(_innerEllipse.radiusMinor(), binY) + ',' +
		    Selecting::convertDimension(_outerEllipse.radiusMajor(), binX) + ',' +
		    Selecting::convertDimension(_outerEllipse.radiusMinor(), binY) + ',' +
		    ::toString<float>(_innerEllipse.rotAngle()) + ',' +
		    ::toString<float>(_outerEllipse.rotAngle()) + ')';
} // ElliptAnnulus<float>::toSelectExpression()
template<class T>
std::string RotBox<T>::toSelectExpression(const std::string &xAxisName,
											  const std::string &yAxisName,
											  double binX, double binY,
											  double minX, double minY) const
	throw(UnexpressionableRegion)
{
	return Selecting::prefix(xAxisName, yAxisName, this->included()) + "box(" +
			Selecting::convertPosition(this->_refPoint.x(), binX, minX,
							this->_refPoint.y(), binY, minY) + ',' +
							Selecting::convertDimension(this->xWidth(), .5 * binX) + ',' +
							Selecting::convertDimension(this->yWidth(), .5 * binY) + ',' +
			::toString(this->rotAngle()) + ')';
} // RotBox<float>::toSelectExpression()

template std::string RotBox<float>::toSelectExpression(const std::string &xAxisName,
											  const std::string &yAxisName,
											  double binX, double binY,
											  double minX, double minY) const
	throw(UnexpressionableRegion);

template<class T>
std::string Polygon<T>::toSelectExpression(const std::string &xAxisName,
										   const std::string &yAxisName,
										   double binX, double binY,
										   double minX, double minY) const
	throw(UnexpressionableRegion)
{
	std::string res;
	unsigned N = _points.size();
	for (unsigned n=0; n<N; ++n)
		res += Selecting::convertPosition(_points[n].x(), binX, minX,
							   _points[n].y(), binY, minY) + ',';

	return Selecting::prefix(xAxisName, yAxisName, this->included()) + "polygon(" +
		   res.substr(0, res.length()-1) + ')';
} // Polygon<T>::toSelectExpression()

template std::string Polygon<float>::toSelectExpression(const std::string &,
				   const std::string &, double, double, double, double) const
	throw(UnexpressionableRegion);
template std::string Polygon<double>::toSelectExpression(const std::string &,
				   const std::string &, double, double, double, double) const
	throw(UnexpressionableRegion);


namespace {

//
//	function that has the same ordering properties as the function
//	that computes the angle subtended by a line from points p2 to p1
//	and p1 to the positive horizontal direction
//	see "Algorithms in C", Robert Sedgewick, p. 352 for details
//
template<class T>
T theta(const Point<T> &p1, const Point<T> &p2) {
	T dx = p2.x() - p1.x(),
	  ax = std::fabs(dx),
	  dy = p2.y() - p1.y(),
	  ay = std::fabs(dy);

	T t = (ax+ay == T(0)) ? 0 : dy/(ax+ay);
	if (dx < 0)
		t = 2-t;
	else if (dy < 0)
		t = 4+t;

	return 90*t;
} // theta()

//
//	function to check if in a path from point A-B-C one turns clock-
//	or counter-clockwise; see "Algorithms in C", Robert Sedgewick, p. 352
//	for details
//
template<class T>
int ccw(const Point<T> &p0,
		const Point<T> &p1,
		const Point<T> &p2)
{
	double dx1, dx2, dy1, dy2;

	dx1 = p1.x() - p0.x(); dy1 = p1.y() - p0.y();
	dx2 = p2.x() - p0.x(); dy2 = p2.y() - p0.y();
	if (dx1*dy2 >= dy1*dx2)						return +1;
	if (dx1*dy2 <= dy1*dx2)						return -1;
	if ((dx1*dx2 <= 0) || (dy1*dy2 <= 0))		return -1;
	if ((dx1*dx1+dy1*dy1) <= (dx2*dx2+dy2*dy2))	return +1;

	return 0;
} // ccw()

//
//	function to check if two line segments intersect;
//	see "Algorithms in C", Robert Sedgewick, p. 352 for details
//
template<class T>
bool intersect(const Point<T> &l1p1, const Point<T> &l1p2,
			   const Point<T> &l2p1, const Point<T> &l2p2)
{
	return ((ccw(l1p1, l1p2, l2p1)
			*ccw(l1p1, l1p2, l2p2)) <= 0)
	  &&   ((ccw(l2p1, l2p2, l1p1)
			*ccw(l2p1, l2p2, l1p2)) <= 0);
} // intersect()

}


template<class T>
class AnchoredPointList : public std::vector<pair<T, Point<T> > > {
public:
	typedef pair<T, Point<T> > Pair;
	AnchoredPointList(const PointList<T> &pl) {
		//gcc 4.8.1 migration: adding "this->"
		this->push_back(Pair(0., pl[0]));
		unsigned N = pl.size();
		for (unsigned n=1; n<N; ++n)
			this->push_back(Pair(theta(pl[n], pl[0]), pl[n])); //gcc 4.8.1 migration: adding "this->"

	}

	bool operator()(const Pair &p1, const Pair &p2) {
		return p1.first < p2.first;
	}

	PointList<T> pointList() const {
		unsigned N = this->size();
		PointList<T> pl;
		for (unsigned n=0; n<N; ++n)
			pl.push_back((*this)[n].second);

		return pl;
	}
};

template<class T>
void PointList<T>::simpleClosedPath()
{
	if (_simpleClosedPathOrder)
		return;	// already ordered

	//	swap first element with the one with smallest x coordinate
	typename PointList<T>::iterator me =
							std::min_element(this->begin(), this->end());
	std::iter_swap(this->begin(), me);

	AnchoredPointList<T> apl(*this);
	std::sort(apl.begin()+1, apl.end(), apl);
	std::vector<Point<T> > &v0 = *this,
						   v1 = apl.pointList();
	v0.swap(v1);

	_simpleClosedPathOrder = true;

	SAS_DEBUG(std::cout<<"Simple-closed-path point list: "<<*this<<endl;)
} // PointList<T>::simpleClosedPath()

template void PointList<float>::simpleClosedPath();
template void PointList<double>::simpleClosedPath();

template<class T>
void PointList<T>::antiClockWiseOrder()
{
	if (_antiClockWiseOrder)
		return;	// already ordered

	//	rotate vector such that first point has smallest X/Y coordinate
	typename PointList<T>::iterator me = std::min_element(this->begin(), this->end());
	std::rotate(this->begin(), me, this->end());

	//	now check the orientation - if it is clockwise, reverse
	std::vector<Point<T> > &pl = *this;
	if (theta(pl[1], pl[0]) > theta(pl.back(), pl[0]))
		std::reverse(pl.begin()+1, pl.end());

	_antiClockWiseOrder = true;

	SAS_DEBUG(std::cout<<"Anti-clockwise-order point list: "<<*this<<endl;)
} // PointList<T>::antiClockWiseOrder()

template void PointList<float>::antiClockWiseOrder();
template void PointList<double>::antiClockWiseOrder();

template<class T>
bool PointList<T>::hasIntersectingSegments() const
{
	if (_simpleClosedPathOrder)
		//	cannot have intersecting segments in this case
		return false;

	const vector<Point<T> > &pl = *this;
	unsigned N = this->size();
	for (unsigned i=0; i<N-1; ++i)
		for (unsigned j=i+2; j<N; ++j)
			if (intersect(pl[i], pl[i+1<N ? i+1 : 0],
						  pl[j], pl[j+1<N ? j+1 : 0]))
				return true;

	return false;
} // PointList<T>::hasIntersectingSegments()

template bool PointList<float>::hasIntersectingSegments() const;
template bool PointList<double>::hasIntersectingSegments() const;

template<class T>
bool PointList<T>::formsSimpleRectangle()
{
	simpleClosedPath();

	const std::vector<Point<T> > &pl = *this;

	if (pl.size() != 4)
		return false;

	return pl[0].y == pl[1].y &&
		   pl[1].x == pl[2].x &&
		   pl[2].y == pl[3].y &&
		   pl[3].x == pl[0].x;
} // PointList<T>::formsSimpleRectangle()
